local GIFTSHOWER_TIMERNAME = "giftshower_spawntimer"
local _worldsettingstimer = GetWorld().components.worldsettingstimer
local _totalgiftshowers = 0
local _cheatedgiftshowers = 0

local function FillAndDropGift(pos, sled)
	for i, items in ipairs(GetWorld().components.giftshowerloot:GetLoot()) do
		for i, v in ipairs(items) do
			if type(v) == "string" then
				items[i] = SpawnPrefab(v)
			else
				items[i] = SpawnPrefab(v[1])
				items[i].components.stackable.stacksize = v[2]
			end
		end

		local theta = math.random() * 2 * PI
		local radius = math.random(3,4)
		local offset = Vector3(radius * math.cos( theta ), 0, -radius * math.sin( theta ))

		local gift = SpawnPrefab("gift")
		gift.components.unwrappable:WrapItems(items)
		for i, v in ipairs(items) do
			v:Remove()
		end
		gift.Transform:SetPosition(pos.x + offset.x, pos.y + math.random(12, 18) + offset.y, pos.z + offset.z)
		gift.status = GetWorld().components.giftshowerloot:GetStatus()
	end
	if GetWorld().components.giftshowerloot:GetStatus() == "nice" then
		GetWorld().SoundEmitter:PlaySound("dontstarve/sled/laughter", nil, .5)
	end
	GetWorld().components.giftshowerloot:NilifyLoot()
	GetWorld().components.giftshowerloot:SetStatus(nil)
	sled:PushEvent("stop_movement")
end

local function SentSled()
    local player_pt = Vector3(GetPlayer().Transform:GetWorldPosition())
    local sledSpawnDistance_x = 20
    local sledSpawnPos = player_pt + Vector3(sledSpawnDistance_x, 0, 0)

    local sled = SpawnPrefab("sled")
	sled.spawnedin = false
    sled.Transform:SetPosition(sledSpawnPos.x, sledSpawnPos.y, sledSpawnPos.z)
	sled:DoTaskInTime(2.5, function(inst)
		FillAndDropGift(sled:GetPosition(), sled)
	end)
end

local function StopRespawnTimer()
    _worldsettingstimer:StopTimer(GIFTSHOWER_TIMERNAME)
end

local function StartRespawnTimer(t)
	StopRespawnTimer()
	_worldsettingstimer:StartTimer(GIFTSHOWER_TIMERNAME, t)
end

local function StartShowerTimer(delay)
    StartRespawnTimer((delay or 0) + (TUNING.GIFTSHOWER_SPAWN_DELAY + math.random() * TUNING.GIFTSHOWER_SPAWN_DELAY_VARIANCE))
end

local function OnSeasonChange(inst, data)
	if SaveGameIndex:GetCurrentMode() ~= "adventure" then
		if data.season == SEASONS.WINTER then
			print("GiftShowerSpawner: It Is Winter!")
			if not _worldsettingstimer:ActiveTimerExists(GIFTSHOWER_TIMERNAME) then
				StartShowerTimer(TUNING.GIFTSHOWER_TURNED_WINTER_DELAY)
			end
		else
			print("GiftShowerSpawner: Winter Has Ended!")
			StopRespawnTimer()
		end
	end
end

local function OnRespawnTimer()
	if GetPlayer() ~= nil and GetPlayer():IsValid() then
		print("GiftShowerSpawner: Spawning Gift Shower!")
		SentSled()
		_totalgiftshowers = _totalgiftshowers + 1
		StartShowerTimer(TUNING.GIFTSHOWER_RESPAWN_DELAY)
	end
end

local function ForceGiftShower()
	SentSled()
	_cheatedgiftshowers = _cheatedgiftshowers + 1
end

local GiftShowerSpawner = Class(function(self, inst)
	self.inst = inst

	_worldsettingstimer:AddTimer(GIFTSHOWER_TIMERNAME, TUNING.GIFTSHOWER_RESPAWN_DELAY + TUNING.GIFTSHOWER_SPAWN_DELAY + TUNING.GIFTSHOWER_SPAWN_DELAY_VARIANCE, true, OnRespawnTimer)

	self.inst:ListenForEvent("seasonChange", OnSeasonChange)
	self.inst:ListenForEvent("forcegiftshower", ForceGiftShower)
end)

function GiftShowerSpawner:OnSave()
    return
    {
        totalgiftshowers = _totalgiftshowers,
		cheatedgiftshowers = _cheatedgiftshowers,
    }
end

function GiftShowerSpawner:OnLoad(data)
	if data.totalgiftshowers then
		_totalgiftshowers = data.totalgiftshowers
	end
	if data.cheatedgiftshowers then
		_cheatedgiftshowers = data.cheatedgiftshowers
	end
end

function GiftShowerSpawner:GetDebugString()
    local time_remaining = _worldsettingstimer:GetTimeLeft(GIFTSHOWER_TIMERNAME) or 0
	if not GetSeasonManager():IsWinter() then
		return "Waiting for winter."
	else
		return string.format("Total Gift Showers: %d\nCheated Gift Showers: %d\nShower Occurring in %.2f (%.2f days)", _totalgiftshowers, _cheatedgiftshowers, time_remaining, time_remaining / TUNING.TOTAL_DAY_TIME)
    end
end

return GiftShowerSpawner
