local function OnStartBundling(inst)
    inst.components.stackable:Get():Remove()
end

local function MakeWrap(name, containerprefab, tag, cheapfuel)
    local assets =
    {
        Asset("ANIM", "anim/"..name..".zip"),
        Asset("INV_IMAGE", "bundlewrap"),
    }

    local prefabs =
    {
        name,
        containerprefab,
    }

    local function fn()
        local inst = CreateEntity()

        inst.entity:AddTransform()

        MakeInventoryPhysics(inst)

        inst.entity:AddAnimState()
        inst.AnimState:SetBank(name)
        inst.AnimState:SetBuild(name)
        inst.AnimState:PlayAnimation("idle")

        if tag ~= nil then
            inst:AddTag(tag)
        end

        inst:AddComponent("stackable")
        inst.components.stackable.maxsize = TUNING.STACK_SIZE_MEDITEM

        inst:AddComponent("inspectable")
        inst:AddComponent("inventoryitem")

        inst:AddComponent("bundlemaker")
        inst.components.bundlemaker:SetBundlingPrefabs(containerprefab, name)
        inst.components.bundlemaker:SetOnStartBundlingFn(OnStartBundling)

        inst:AddComponent("fuel")
        inst.components.fuel.fuelvalue = cheapfuel and TUNING.TINY_FUEL or TUNING.MED_FUEL

        MakeSmallBurnable(inst, TUNING.SMALL_BURNTIME)
        MakeSmallPropagator(inst)
        inst.components.propagator.flashpoint = 10 + math.random() * 5

		if KnownModIndex:IsModEnabled("workshop-3063944835") then
			inst.components.inventoryitem.nosink = true

			inst:AddComponent("modfloatable")
			inst.components.modfloatable.heavyitem = false
		end

        return inst
    end

    return Prefab(name.."wrap", fn, assets, prefabs)
end

local function CollectContainerItems(container)
    local items = {}

    for i = 1, container:GetNumSlots() do
        local item = container:GetItemInSlot(i)
        if item ~= nil then
            table.insert(items, item)
        end
    end

    return items
end

local function ItemTestFn(inst, item, slot)
    return not (item:HasTag("irreplaceable") or item:HasTag("bundle") or item:HasTag("nobundling"))
end

local slotpos = {
    Vector3(-37.5, 32 + 4, 0), 
    Vector3(37.5, 32 + 4, 0),
    Vector3(-37.5, -(32 + 4), 0), 
    Vector3(37.5, -(32 + 4), 0),
}

local function onburnt(inst)
    inst.burnt = true
    inst.components.unwrappable:Unwrap()
end

local function onignite(inst)
    inst.components.unwrappable.canbeunwrapped = false
end

local function onextinguish(inst)
    inst.components.unwrappable.canbeunwrapped = true
end

local function MakeBundle(name, onesize, variations, loot, tossloot, setupdata)
    local assets =
    {
        Asset("ANIM", "anim/"..name..".zip"),
    }

    if variations ~= nil then
        for i = 1, variations do
            if onesize then
                table.insert(assets, Asset("INV_IMAGE", name..tostring(i)))
            else
                table.insert(assets, Asset("INV_IMAGE", name.."_small"..tostring(i)))
                table.insert(assets, Asset("INV_IMAGE", name.."_medium"..tostring(i)))
                table.insert(assets, Asset("INV_IMAGE", name.."_large"..tostring(i)))
            end
        end
    elseif not onesize then
        table.insert(assets, Asset("INV_IMAGE", name.."_small"))
        table.insert(assets, Asset("INV_IMAGE", name.."_medium"))
        table.insert(assets, Asset("INV_IMAGE", name.."_large"))
    end

    local prefabs =
    {
        "ash",
        name.."_unwrap",
    }

    if loot ~= nil then
        for i, v in ipairs(loot) do
            table.insert(prefabs, v)
        end
    end

    local function OnWrapped(inst, num, doer)
        local suffix =
            (onesize and "_onesize") or
            (num > 3 and "_large") or
            (num > 1 and "_medium") or
            "_small"

        if variations ~= nil then
            if inst.variation == nil then
                inst.variation = math.random(variations)
            end
            suffix = suffix..tostring(inst.variation)
            inst.components.inventoryitem:ChangeImageName(name..(onesize and tostring(inst.variation) or suffix))
        elseif not onesize then
            inst.components.inventoryitem:ChangeImageName(name..suffix)
        end

        inst.AnimState:PlayAnimation("idle"..suffix)
    --	inst.components.floatable:UpdateAnimations("idle"..suffix.."_water", "idle"..suffix)

        if doer ~= nil and doer.SoundEmitter ~= nil then
            doer.SoundEmitter:PlaySound("dontstarve/common/craftable/bundles/packaged")
        end
    end

    local function OnUnwrapped(inst, pos, doer)
        if inst.burnt then
            SpawnPrefab("ash").Transform:SetPosition(pos:Get())
        else
            local loottable = (setupdata ~= nil and setupdata.lootfn ~= nil) and setupdata.lootfn(inst, doer) or loot
            if loottable ~= nil then
                for i, v in ipairs(loottable) do
                    local item = SpawnPrefab(v)
                    if item ~= nil then
                        item.Transform:SetPosition(pos:Get())
						
						if item.components.inventoryitem then
							item.components.inventoryitem:OnDropped(true, nil, nil, .5)
						end

                        inst:ApplyInheritedMoisture(item)
                    end
                end
            end

            SpawnPrefab(name.."_unwrap").Transform:SetPosition(pos:Get())
        end

		if inst.status ~= nil then
			if doer ~= nil and doer.components.sanity then
				if inst.status == "nice" then
					doer.components.sanity:DoDelta(TUNING.SANITY_MEDLARGE)
				elseif inst.status == "moderate" then
					doer.components.sanity:DoDelta(TUNING.SANITY_SMALL)
				end
			end

			if doer ~= nil and inst.SoundEmitter ~= nil then
				inst.SoundEmitter:PlaySound("dontstarve/giftshower/"..inst.status)
			end
		end

        if doer ~= nil and doer.SoundEmitter ~= nil then
            doer.SoundEmitter:PlaySound("dontstarve/common/craftable/bundles/packaged")
        end
        inst:Remove()
    end

    local OnSave = variations ~= nil and function(inst, data)
        data.variation = inst.variation
		data.status = inst.status
    end or nil

    -- TODO: pre load???
    local OnPreLoad = variations ~= nil and function(inst, data)
        if data ~= nil then
            inst.variation = data.variation
			inst.status = data.status
        end
    end or nil

    local function fn()
        local inst = CreateEntity()

        inst.entity:AddTransform()
        inst.entity:AddAnimState()
		inst.entity:AddSoundEmitter()

        MakeInventoryPhysics(inst)

        inst.AnimState:SetBank(name)
        inst.AnimState:SetBuild(name)
        inst.AnimState:PlayAnimation(
            variations ~= nil and
            (onesize and "idle_onesize1" or "idle_large1") or
            (onesize and "idle_onesize" or "idle_large")
        )

        inst:AddTag("bundle")
        inst:AddTag("unwrappable")

		inst.status = nil

        if setupdata ~= nil and setupdata.common_postinit ~= nil then
            setupdata.common_postinit(inst, setupdata)
        end

		inst:AddComponent("inspectable")
		inst.components.inspectable.getstatus = function(inst)
			if inst.status == "nice" then 
				return "NICE"
			elseif inst.status == "moderate" then
				return "MODERATE"
			elseif inst.status == "naughty" then
				return "NAUGHTY"
			else
				return "GENERIC"
			end
		end

        inst:AddComponent("inventoryitem")

        if variations ~= nil or not onesize then
            inst.components.inventoryitem:ChangeImageName(
                name..
                (variations == nil and "_large" or (onesize and "1" or "_large1"))
            )
        end

        inst:AddComponent("unwrappable")
        inst.components.unwrappable:SetOnWrappedFn(OnWrapped)
        inst.components.unwrappable:SetOnUnwrappedFn(OnUnwrapped)

        MakeSmallBurnable(inst, TUNING.SMALL_BURNTIME)
        MakeSmallPropagator(inst)
        inst.components.propagator.flashpoint = 10 + math.random() * 5
        inst.components.burnable:SetOnBurntFn(onburnt)
        inst.components.burnable:SetOnIgniteFn(onignite)
        inst.components.burnable:SetOnExtinguishFn(onextinguish)

        if setupdata ~= nil and setupdata.master_postinit ~= nil then
            setupdata.master_postinit(inst, setupdata)
        end

        inst:AddComponent("useableitem")
        inst.components.useableitem:SetOnUseFn(function ()
            inst.components.unwrappable:Unwrap()
        end)
        inst.components.useableitem:SetCanInteractFn(function() return true end)

        inst.OnSave = OnSave
        inst.OnPreLoad = OnPreLoad

		if KnownModIndex:IsModEnabled("workshop-3063944835") then
			inst.components.inventoryitem.nosink = true

			inst:AddComponent("modfloatable")
			inst.components.modfloatable.heavyitem = false
			inst.components.modfloatable.size = "small"
		end

        return inst
    end

    return Prefab(name, fn, assets, prefabs)
end

local function MakeFX(name)
    local assets =
    {
        Asset("ANIM", "anim/"..name..".zip"),
    }

	local function fn()
		local inst = CreateEntity()
		inst.entity:AddTransform()
		inst.entity:AddAnimState()
		inst.entity:AddSoundEmitter()
		inst.persists = false
		
		inst.AnimState:SetBank("gift")
		inst.AnimState:SetBuild("gift")
		inst.AnimState:PlayAnimation("unwrap")

		inst:AddTag("FX")
		inst:AddTag("NOCLICK")
		inst:AddTag("NOBLOCK")

		return inst
    end

    return Prefab(name.."_unwrap", fn, assets)
end

return MakeWrap("gift", "bundle_container", "gift", true),
		MakeBundle("gift", false, 2),
		MakeFX("gift")